from collections.abc import MutableMapping
from typing import Any, cast


def update_nested(d: MutableMapping[str, Any], u: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
    """Update a nested dictionary or similar mapping.
    This is a recursive function that will update the values in the dictionary
    *inplace*.

    Parameters
    ----------
    d : MutableMapping
        The dictionary to update.
    u : MutableMapping
        The dictionary to update from.

    Returns
    -------
    MutableMapping
        The updated dictionary (same as d).
    """
    for k, v in u.items():
        if isinstance(v, MutableMapping):
            d[k] = update_nested(d.get(k, {}), cast(MutableMapping[str, Any], v))
        else:
            d[k] = v
    return d



"""Remote Demanders mechanisms

This module contains the different demanders mechanisms
that can be used to demand a TOKEN from a previously
discovered endpoint.


These are NOT grants, and are used by the remote grants
to discover the endpoint to connect to.

"""



"""Remote Grants

Fakts remote grants are used to retrieve configuration from a remote endpoint.
All this grants are based on the Fakts Registration Protocol that tries to ensure
legitmitate registration of apps on a dynamic endpoint



"""

from .base import RemoteGrant
from .models import Demander, Discovery, FaktsEndpoint, Claimer


__all__ = ["RemoteGrant", "Demander", "Discovery", "FaktsEndpoint", "Claimer"]



import aiohttp
from typing import Optional
from pydantic import BaseModel, ConfigDict, Field
import logging
from fakts_next.grants.remote.errors import DemandError
from fakts_next.grants.remote.models import FaktsEndpoint
import ssl
import certifi

logger = logging.getLogger(__name__)


class RetrieveError(DemandError):
    """A base class for all retrieve errors"""

    pass


class RetrieveDemander(BaseModel):
    """Retrieve Demander

    A retrieve grant is a remote grant can be used to retrieve a token and a configuration from a fakts_next server, by claiming to be an already
    registed public application on the fakts_next server. Public applications are applications that are not able to keep a secret, and therefore
    need users to explicitly grant them access to their data. YOu need to also provide a redirect_uri that matches the one that is registered
    on the fakts_next server.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    manifest: BaseModel
    """ The manifest of the application that is requesting the token"""

    retrieve_url: Optional[str] = Field(
        None,
        description="The url to use for retrieving the token (overwrited the endpoint url)",
    )
    """The url to use for retrieving the token (overwrited the endpoint url)"""

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        """Demand a token from the endpoint

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to demand the token from
        request : FaktsRequest
            The request to use for the demand

        Returns
        -------
        str
            The token that was retrieved
        """

        retrieve_url = self.retrieve_url or endpoint.retrieve_url or f"{endpoint.base_url}retrieve/"

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl_context)) as session:
            logger.debug(f"Requesting token from {retrieve_url}")
            async with session.post(
                retrieve_url,
                json={
                    "manifest": self.manifest.model_dump(),
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if "status" not in data:
                        raise RetrieveError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise RetrieveError(data["message"])
                    if status == "granted":
                        return data["token"]

                    raise RetrieveError(f"Unexpected status: {status}")
                else:
                    raise RetrieveError("Error! Coud not claim this app on this endpoint")

    async def arefresh(self, endpoint: FaktsEndpoint) -> str:
        """Refreshes the token for the given endpoint.

        This method will refresh the token for the given endpoint. This method will
        request a new code from the fakts_next server. This code will be used to
        authenticate the user. The user will be prompted to visit a URL and enter the code.

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to fetch the token for
        request : FaktsRequest
            The request to use for the fetching of the token

        Returns
        -------
        str
            The token that was refreshed
        """

        return await self.ademand(endpoint)



from fakts_next.fakts import get_current_fakts_next
from fakts_next.models import Alias
from koil.helpers import unkoil


async def afakt(key: str, omit_challenge: bool | None = None) -> Alias:
    value = await get_current_fakts_next().aget_alias(
        key, omit_challenge=omit_challenge
    )
    return value


def fakt(key: str, omit_challenge: bool | None = None) -> Alias:
    return unkoil(afakt, key, omit_challenge=omit_challenge)



class FaktsError(Exception):
    """Base class for all Fakts errors

    This class is used to catch all Fakts errors. If you want to catch
    all Fakts errors, you can catch this class.
    """


class NoFaktsFound(FaktsError):
    """Raised when no fakts_next instance is found in the current context.

    If this error is raised, it means that you are trying to access
    the fakts_next instance from a context where it is not available. Online
    places where it is available are:

    ```python

    with Fakts(grant=grant) as fakts_next:
        # fakts_next is available here

    async with Fakts(grant=grant) as fakts_next:
        # fakts_next is available here

    fake_fakts_next = Fakts(grant=grant)
    # fakt is not available here

    ```


    """


class GroupNotFound(FaktsError):
    """Raised when a group is not found in the configuration

    This error is raised when a group is not found in the configuration.
    This can happen when you try to access a group that does not exist
    in the active configuration. Also, this can be raised after a reload.




    """



"""Fakts grants module.

This module contains the grants that are included with
the fakts_next package.

Grants are the way to get configuration from different
sources. They are the main way to get configuration, and
are used by the Fakts class to get the configuration.

Generally, the grants are split into a few categories:

- Remote grants: These grants are used to connect to
    remote configuration servers, and fetch the configuration
    from there. They are generally used to fetch dynamic
    configuration from a remote server.
- Hard grants: These grants are used to
    hardcode the configuration into the code. They are
    generally used for testing purposes, or to provide
    a default configuration that can be used if no other
    configuration is available.

"""

from .env import EnvGrant
from .errors import GrantError
from .remote import RemoteGrant


__all__ = [
    "EnvGrant",
    "GrantError",
    "RemoteGrant",
]



from .base import BaseOauth2Grant
import logging


logger = logging.getLogger(__name__)


def build_authorize_url(grant: BaseOauth2Grant) -> str:
    """Builds the authorize url for the given grant.

    Parameters
    ----------
    grant : BaseOauth2Grant
        A BaseOauth2Grant

    Returns
    -------
    str
        The authorize url
    """
    return f"{grant.base_url}/{grant.authorize_path}/" if grant.append_trailing_slash else f"{grant.base_url}/{grant.authorize_path}"


def build_token_url(grant: BaseOauth2Grant) -> str:
    """Builds the token url for the given grant.

    Parameters
    ----------
    grant : BaseOauth2Grant
        BaseOauth2Grant

    Returns
    -------
    str
        The token url
    """
    return f"{grant.base_url}/{grant.token_path}/" if grant.append_trailing_slash else f"{grant.base_url}/{grant.token_path}"


def build_refresh_url(grant: BaseOauth2Grant) -> str:
    """Builds the token url for the given grant.

    Parameters
    ----------
    grant : BaseOauth2Grant
        BaseOauth2Grant

    Returns
    -------
    str
        The token url
    """
    return f"{grant.base_url}/{grant.refresh_path}/" if grant.append_trailing_slash else f"{grant.base_url}/{grant.refresh_path}"



from pydantic import BaseModel
from fakts_next.grants.remote.models import FaktsEndpoint


class StaticDiscovery(BaseModel):
    """A discovery that always returns the same endpoint

    This is mostly used for testing purposes.
    """

    endpoint: FaktsEndpoint

    async def adiscover(self) -> FaktsEndpoint:
        """Discover the endpoint

        This method will always return the same endpoint (the one that was
        passed to the constructor)

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the discovery process (is not used)

        Returns
        -------
        FaktsEndpoint
            A valid endpoint
        """

        return self.endpoint



from typing import Dict, AsyncGenerator, List, Tuple, Set
from pydantic import ConfigDict, Field, field_validator
from socket import AF_INET, IPPROTO_UDP
import asyncio
import json
import logging
from pydantic import BaseModel
import ssl
import certifi
from .utils import discover_url
from fakts_next.grants.remote.models import FaktsEndpoint
from fakts_next.grants.remote.errors import DiscoveryError

logger = logging.getLogger(__name__)

DataGram = Tuple[bytes, Tuple[str, int]]


class DiscoveryProtocol(asyncio.DatagramProtocol):
    "The protocol that is used to receive beacons, and put them in a queue"

    def __init__(self, recvq: asyncio.Queue[DataGram]) -> None:
        """Initialize the protocol

        Parameters
        ----------
        recvq : asyncio.Queue
            The queue to put the beacons in
        """
        super().__init__()
        self._recvq = recvq

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Receive a datagram

        This method is called when a datagram is received, and
        puts it in the queue

        Parameters
        ----------
        data : bytes
            The data that was received
        addr : Tuple[str, int]
            The address it was received from
        """
        self._recvq.put_nowait((data, addr))


class ListenBinding(BaseModel):
    """A binding to listen on for beacons"""

    address: str = "0.0.0.0"
    port: int = 45678
    magic_phrase: str = "beacon-fakts_next"

    @field_validator("port")
    def check_port(cls, v: int) -> int:
        if v < 0 or v > 65535:
            raise ValueError(f"Port {v} is not in the valid range")
        return v


class Beacon(BaseModel):
    """A beacon that is received when listening on
    a broadcast port"""

    url: str
    """The url of the endpoint"""


async def alisten(bind: ListenBinding, strict: bool = False) -> AsyncGenerator[Beacon, None]:
    """A generator that listens on a broadcast port for beacons

    This generator listens on a specific binding for beacons.
    It will yield the beacons as it receives.


    Parameters
    ----------
    bind : ListenBinding
        The binding to listen on
    strict : bool, optional
        Should we error on bad Beacons, by default False


    Yields
    ------
    Beacon
        The beacon that was received

    Raises
    ------
    e
        Any exception that is raised by the socket
    """
    transport = None

    try:
        loop = asyncio.get_event_loop()
        read_queue: asyncio.Queue[DataGram] = asyncio.Queue()  # type: ignore
        transport, _ = await loop.create_datagram_endpoint(
            lambda: DiscoveryProtocol(read_queue),
            local_addr=(bind.address, bind.port),  # Change port number here
            family=AF_INET,
            proto=IPPROTO_UDP,
        )

        while True:
            data, _ = await read_queue.get()
            try:
                data = str(data, "utf8")
                if data.startswith(bind.magic_phrase):
                    endpoint = data[len(bind.magic_phrase) :]

                    try:
                        endpoint = json.loads(endpoint)
                        endpoint = Beacon(**endpoint)
                        yield endpoint

                    except json.JSONDecodeError as e:
                        logger.error("Received Request but it was not valid json")
                        if strict:
                            raise e

                else:
                    logger.error(f"Received Non Magic Response {data}. Maybe somebody sends")

            except UnicodeEncodeError as e:
                logger.error("Couldn't decode received message")
                if strict:
                    raise e

    except asyncio.CancelledError as e:
        if transport:
            transport.close()
        logger.info("Stopped checking")
        raise e
    finally:
        if transport:
            transport.close()
        logger.info("Stopped checking")


async def alisten_pure(bind: ListenBinding, strict: bool = False) -> AsyncGenerator[Beacon, None]:
    """A generator that listens on a broadcast port for beacons

    This generator listens on a specific binding for beacons.
    It will yield the beacons as it receives, but will only yield
    each beacon once.


    Parameters
    ----------
    bind : ListenBinding
        The binding to listen on
    strict : bool, optional
        Should we error on bad Beacons, by default False


    Yields
    ------
    Beacon
        The beacon that was received

    Raises
    ------
    e
        Any exception that is raised by the socket
    """

    already_detected: Set[str] = set()

    async for x in alisten(bind, strict):
        if x.url not in already_detected:
            already_detected.add(x.url)
            yield x

    return


class FirstAdvertisedDiscovery(BaseModel):
    """A discovery that will return the first endpoint that is advertised

    This discovery will listen on a broadcast port for beacons.
    It will then try to connect to the endpoint and return it.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    binding: ListenBinding = Field(default_factory=ListenBinding)
    """The address to bind to"""
    strict: bool = False
    """Should we error on bad Beacons"""
    discovered_endpoints: Dict[str, FaktsEndpoint] = Field(default_factory=dict)
    """A cache of discovered endpoints"""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )

    async def adiscover(self) -> FaktsEndpoint:
        """Discover the endpoint

        This method will always return the same endpoint (the one that was
        passed to the constructor)

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the discovery process (is not used)

        Returns
        -------
        FaktsEndpoint
            A valid endpoint
        """

        async for beacon in alisten_pure(self.binding, strict=self.strict):
            try:
                endpoint = await discover_url(beacon.url, self.ssl_context)
                return endpoint
            except Exception as e:
                logger.error(f"Could not connect to beacon {beacon.url}: {e}")
                continue

        raise DiscoveryError("Could not find any endpoint")



import logging
from qtpy import QtCore


from fakts_next.grants.remote.models import FaktsEndpoint
from typing import Optional, Dict
import datetime
from fakts_next.protocols import FaktValue
from pydantic import BaseModel, ConfigDict, Field
from fakts_next.cache.model import CacheModel

logger = logging.getLogger(__name__)


class QtSettingsCache(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    settings: QtCore.QSettings  # type: ignore #
    save_key: str = "fakts_cache"
    hash: str = Field(
        default_factory=lambda: "",
        description="Validating against the hash of the config",
    )

    async def aset(self, value: Dict[str, FaktValue]) -> None:
        """Stores the value in the settings

        Parameters
        ----------
        value : Dict[str, FaktValue]
            The value to store
        """

        cache = CacheModel(config=value, created=datetime.datetime.now(), hash=self.hash)

        self.settings.setValue(self.save_key, cache.model_dump_json())  # type: ignore #

    async def aload(self) -> Optional[Dict[str, FaktValue]]:
        """Loads the value from the settings

        Returns
        -------
        Optional[Dict[str, FaktValue]]
            The value, or None if there is no value
        """

        un_storage: str = self.settings.value(self.save_key, None)  # type: ignore #
        if not un_storage:
            return None

        if not isinstance(un_storage, str):
            logger.warning("Cache is not a string")
            raise ValueError("Cache is not a string")
        try:
            storage = CacheModel.model_validate_json(un_storage)
            if storage.hash != self.hash:
                return None

            return storage.config
        except Exception as e:
            logger.error("Cache is not a string", exc_info=e)

        return None

    async def areset(self) -> Optional[FaktsEndpoint]:
        """A function that gets the default endpoint

        Returns
        -------
        Optional[FaktsEndpoint]
            The stored endpoint, or None if there is no endpoint

        """

        self.settings.setValue(self.save_key, None)  # type: ignore #



from qtpy import QtWidgets
from koil.qt import qt_to_async  # type: ignore
from fakts_next.grants.remote.models import FaktsEndpoint


class ShouldWeSaveDialog(QtWidgets.QDialog):  # type: ignore
    """A dialog that asks the user if we should save the token or not"""

    def __init__(self, endpoint: FaktsEndpoint, token: str, *args, **kwargs) -> None:  # type: ignore
        """Constructor for ShouldWeSaveDialog"""
        super().__init__(*args, **kwargs)  # type: ignore
        self.setWindowTitle(f"Connected to {endpoint.name}")  # type: ignore

        self.qlabel = QtWidgets.QLabel("Do you want to auto save this configuration for this endpoint?")

        vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(vlayout)  # type: ignore

        vlayout.addWidget(self.qlabel)

        hlayout = QtWidgets.QHBoxLayout()
        vlayout.addLayout(hlayout)

        self.yes_button = QtWidgets.QPushButton("Yes")
        self.no_button = QtWidgets.QPushButton("No")

        self.yes_button.clicked.connect(self._on_yes)
        self.no_button.clicked.connect(self._on_no)

        hlayout.addWidget(self.yes_button)
        hlayout.addWidget(self.no_button)

    def _on_yes(self) -> None:
        self.accept()  # type: ignore

    def _on_no(self) -> None:
        self.reject()  # type: ignore


class AutoSaveTokenWidget(QtWidgets.QWidget):  # type: ignore
    """A simple widget that asks the user if we should save the token or not"""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        """Constructor for AutoSaveTokenWidget"""
        super().__init__(*args, **kwargs)  # type: ignore

        self.ashould_we = qt_to_async(self._should_we, autoresolve=True)  # type: ignore

    def _should_we(self, endpoint: FaktsEndpoint, token: str) -> bool:
        dialog = ShouldWeSaveDialog(endpoint, token, parent=self)
        dialog.exec_()  # type: ignore
        return dialog.result() == QtWidgets.QDialog.Accepted  # type: ignore

    async def ashould_we_save(self, endpoint: FaktsEndpoint, token: str) -> bool:
        """Should ask the user if we should save the user"""
        return await self.ashould_we(endpoint, token)  # type: ignore



from fakts_next.grants.remote.models import FaktsEndpoint
from pydantic import BaseModel


class StaticDemander(BaseModel):
    """Static Grant

    A static demander is a remote grant that has a static token. This token can
    for example have been retrieved from a configuration file beforehand and uniquely
    identifies the application on the fakts_next server. When using the static grant make
    sure that the token is not shared with other applications. As they can then mimik
    your application.

    Attention: If you are using the static grant, make sure that the token is not
    shared with other applications. As they can then mimik your application, especially
    when this static token maps to an client-credentials (user) application on the fakts_next
    server, as this application will then be able to access the data of the user that
    granted the application in the first place.

    """

    token: str
    """ The token (secret) that uniquely identifies this application on the fakts_next server."""

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        """Demand a token from the endpoint

        Retrieve the token that was provided to the demander

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to demand the token from

        request : FaktsRequest
            The request to use for the demand

        Returns
        -------
        str
            The token that was retrieved
        """
        return self.token



import aiohttp
from typing import Optional
from pydantic import BaseModel, Field, ConfigDict
import logging
from fakts_next.grants.remote.errors import DemandError
from fakts_next.grants.remote.models import FaktsEndpoint
import ssl
import certifi

logger = logging.getLogger(__name__)


class RetrieveError(DemandError):
    """A base class for all retrieve errors"""

    pass


class RedeemDemander(BaseModel):
    """Redeem Demander

    A reedem grant is a remote grant that can be used to in one shot, create a new client and retrieve a token and a configuration from a fakts_next server.
    The redeem token is a token that was issued by the fakts_next server before, and that can be used to create a any new client (restricted to development
    clients bound to one user). This is useful for creating new clients in an environment where the client CAN keep a secret, but where the clients manifest
    is not known to the fakts_next server.

    """

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    manifest: BaseModel
    """ The manifest of the application that is requesting the token"""

    token: str
    """ The token with which to redeem the client"""

    retrieve_url: Optional[str] = Field(
        default=None,
        description="The url to use for retrieving the token (overwrited the endpoint url)",
    )
    """The url to use for retrieving the token (overwrited the endpoint url)"""
    model_config = ConfigDict(arbitrary_types_allowed=True)

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        """Demand a token from the endpoint

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to demand the token from
        request : FaktsRequest
            The request to use for the demand

        Returns
        -------
        str
            The token that was retrieved
        """

        retrieve_url = self.retrieve_url or endpoint.retrieve_url or f"{endpoint.base_url}redeem/"

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context)
        ) as session:
            logger.debug(f"Requesting token from {retrieve_url}")
            async with session.post(
                retrieve_url,
                json={
                    "manifest": self.manifest.model_dump(),
                    "token": self.token,
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if "status" not in data:
                        raise RetrieveError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise RetrieveError(data["message"])
                    if status == "granted":
                        return data["token"]

                    raise RetrieveError(f"Unexpected status: {status}")
                else:
                    raise RetrieveError("Error! Coud not claim this app on this endpoint")

    async def arefresh(self, endpoint: FaktsEndpoint) -> str:
        """Refreshes the token for the given endpoint.

        This method will refresh the token for the given endpoint. This method will
        request a new code from the fakts_next server. This code will be used to
        authenticate the user. The user will be prompted to visit a URL and enter the code.

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to fetch the token for
        request : FaktsRequest
            The request to use for the fetching of the token

        Returns
        -------
        str
            The token that was refreshed
        """

        return await self.ademand(endpoint)



import logging
from typing import List
import socket
import aiohttp

from fakts_next.grants.remote.models import FaktsEndpoint

logger = logging.getLogger(__name__)


async def acheck_supported_layers(endpoint: FaktsEndpoint) -> List[str]:
    supported_layers: list[str] = []

    if endpoint.layers:
        logger.debug(f"Checking supported layers for {endpoint.name}")
        for layer in endpoint.layers:
            print(f"Checking layer {layer}")
            if layer.get_probe:
                # check if we access to the web layer
                async with aiohttp.ClientSession() as session:
                    try:
                        async with session.get(layer.get_probe) as response:
                            await response.text()

                    except Exception as e:
                        logger.debug(f"Could not access web layer: {e}")
                        continue

            if layer.dns_probe:
                # check if we access to the headscale layer
                try:
                    socket.gethostbyname(layer.dns_probe)
                except Exception as e:
                    print(f"Could not access headscale layer: {e}")
                    logger.debug(f"Could not access headscale layer: {e}")
                    continue

            supported_layers.append(layer.identifier)

    print("Supported layers", supported_layers)

    return supported_layers


def could_copy_to_clipboard(text: str) -> bool:
    """Copies text to clipboard if possible

    This function tries to copy the text to the clipboard.
    If it fails, it returns False, otherwise True.

    Parameters
    ----------
    text : str
        The text to copy to the clipboard

    Returns
    -------
    bool
        Could the text be copied to the clipboard?
    """

    try:
        import pyperclip  # type: ignore[import]

        pyperclip.copy(text)
        return True
    except ImportError:
        logger.debug("Could not import pyperclip, not copying to clipboard")
        return False


try:
    from rich import print as rprint
    from rich.panel import Panel

    def print_device_code_prompt(querystring: str, url: str, code: str) -> None:
        """Prints the device code prompt

        This function prints the device code prompt using rich.
        It also copies the code to the clipboard if possible.

        Parameters
        ----------
        querystring : str
            The querystring to visit
        url : str
            The url to visit (without querystring)
        code : str
            The code to enter on the website
        scopes : List[str]
            The scopes to grant
        """

        could_copy_to_clipboard(code)
        rprint(
            Panel.fit(
                f"""
    Please visit the following URL:
    [bold green][link={querystring}]{querystring}[/link][/bold green]
    or go to this URL:
    [bold green][link={url}]{url}[/link][/bold green]
    and enter the code:
    [bold blue]{code}[/bold blue]
        """,
                title="Device Code Grant",
                title_align="center",
            )
        )

    def print_succesfull_login() -> None:
        """Prints the successful login message

        This function prints the successful login message using rich.
        """
        rprint(
            Panel.fit(
                "You have successfully logged in!",
                title="Device Code Grant",
                title_align="center",
            )
        )

except ImportError:

    def print_device_code_prompt(querystring: str, url: str, code: str) -> None:
        """Prints the device code prompt

        This function prints the device code prompt using rich.
        It also copies the code to the clipboard if possible.

        Parameters
        ----------
        querystring : str
            The querystring to visit
        url : str
            The url to visit (without querystring)
        code : str
            The code to enter on the website
        scopes : List[str]
            The scopes to grant
        """
        could_copy = could_copy_to_clipboard(code)
        print("Please visit the following URL:")
        print("\t" + querystring)
        print("Or go to this URL:")
        print("\t" + url + "device")
        print("And enter the following code:")
        print("\t" + code)
        if could_copy:
            print("Code has been copied to clipboard")

    def print_succesfull_login() -> None:
        """Prints the successful login message

        This function prints the successful login message using rich.
        """
        print("You have successfully logged in!")



from pydantic import BaseModel
from fakts_next.grants.remote import RemoteGrant

from fakts_next.grants.remote.claimers.static import StaticClaimer
from fakts_next.grants.remote.discovery.static import StaticDiscovery
from fakts_next.grants.remote.demanders.static import StaticDemander
from fakts_next.grants.remote.demanders.redeem import RedeemDemander
from fakts_next.grants.remote.claimers.post import ClaimEndpointClaimer
from fakts_next.grants.remote.models import FaktsEndpoint, FaktValue
from typing import Dict


def build_remote_testing(value: Dict[str, FaktValue]) -> RemoteGrant:
    """Builds a remote grant for testing purposes

    Will always return the same value when claiming.

    Parameters
    ----------
    value : Dict[str, FaktValue]
        The value to return when claiming

    Returns
    -------
    RemoteGrant
        The remote grant

    """
    return RemoteGrant(
        discovery=StaticDiscovery(endpoint=FaktsEndpoint(base_url="https://example.com")),
        claimer=StaticClaimer(value=value),
        demander=StaticDemander(token="token"),  # type: ignore
    )


def build_redeem_grant(url: str, manifest: BaseModel, redeem_token: str) -> RemoteGrant:
    return RemoteGrant(
        discovery=StaticDiscovery(endpoint=FaktsEndpoint(base_url=url)),
        claimer=ClaimEndpointClaimer(),
        demander=RedeemDemander(manifest=manifest, token=redeem_token),
    )


def build_remote_testing_with_token(fakts_next_url: str, token: str) -> RemoteGrant:
    """Builds a remote grant for testing purposes

    This grant will use the given token to demand the configuration from fakts_next.
    This is great for testing purposes, or when an api token is known at compile time.

    Parameters
    ----------
    value : Dict[str, FaktValue]
        The value to return when claiming

    Returns
    -------
    RemoteGrant
        The remote grant

    """
    return RemoteGrant(
        discovery=StaticDiscovery(endpoint=FaktsEndpoint(base_url=fakts_next_url)),
        claimer=ClaimEndpointClaimer(),
        demander=StaticDemander(token=token),  # type: ignore
    )



from typing import List, Protocol, runtime_checkable, Optional, Literal
from pydantic import BaseModel
from fakts_next.models import ActiveFakts


class Layer(BaseModel):
    identifier: str
    kind: Literal["WEB", "TAILSCALE"]
    dns_probe: str | None = None
    get_probe: str | None = None
    description: str | None = None


class FaktsEndpoint(BaseModel):
    """FaktsEndpoint

    A FaktsEndpoint is a remote endpoint that can be used to
    retrieve the configuration. This class is used to represent
    the endpoints that are discovered by the discovery mechanisms.
    (For example, when accessing a well-known fakts_next URL)"""

    base_url: str = "http://localhost:8000/f/"
    """The base URL of the endpoint. Akin to the base URL of a Oauth2 """
    name: str = "Helper"
    """ A human readable name for the endpoint"""
    description: Optional[str] = None
    """ A human readable description for the endpoint"""
    retrieve_url: Optional[str] = None
    claim_url: Optional[str] = None
    version: Optional[str] = None
    ca_crt: Optional[str] = None
    layers: Optional[List[Layer]] = None


@runtime_checkable
class Demander(Protocol):
    """A demander takes a FaktsEndpoint and returns the Fakts
    user input.
    """

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        """Demands a token for the given endpoint.

        This method should return the token that can be used to retrieve
        the configuration from the endpoint.

        Args:
            endpoint (FaktsEndpoint): The endpoint to demand the token for.
            request (FaktsRequest): The request that is being processed.

        Returns:
            str: The token that can be used to retrieve the configuration.



        """
        ...


@runtime_checkable
class Discovery(Protocol):
    """Discovery is the abstract base class for discovery mechanisms

    A discovery mechanism is a way to find a Fakts endpoint
    that can be used to retrieve the configuration.

    This class provides an asynchronous interface, as the discovery can
    envolve lenghty operations such as network requests or waiting for
    user input.
    """

    async def adiscover(self) -> FaktsEndpoint:
        """Discovers an endpoint.

        This method should return an endpoint that can be used to retrieve
        the configuration. If no endpoint can be found, it should raise
        a DiscoveryError.

        Parameters
        ----------
        request : FaktsRequest
            The request that is being processed.

        Returns
        -------
        FaktsEndpoint
            The endpoint that can be used to retrieve the configuration.
        """
        ...


@runtime_checkable
class Claimer(Protocol):
    """Discovery is the abstract base class for discovery mechanisms

    A discovery mechanism is a way to find a Fakts endpoint
    that can be used to retrieve the configuration.

    This class provides an asynchronous interface, as the discovery can
    envolve lenghty operations such as network requests or waiting for
    user input.
    """

    async def aclaim(self, token: str, endpoint: FaktsEndpoint) -> ActiveFakts:
        """Discovers an endpoint.

        This method should return an endpoint that can be used to retrieve
        the configuration. If no endpoint can be found, it should raise
        a DiscoveryError.

        Parameters
        ----------
        request : FaktsRequest
            The request that is being processed.

        Returns
        -------
        FaktsEndpoint
            The endpoint that can be used to retrieve the configuration.
        """
        ...



from fakts_next.fakts import Fakts
from rath.links.httpx import HttpxLink
from rath.operation import Operation
from pydantic import BaseModel


class FaltsHttpXConfig(BaseModel):
    """FaltsHttpXConfig"""

    endpoint_url: str


class FaktsHttpXLink(HttpxLink):
    """FaktsHttpXLink


    A FaktsHttpXLink is a HttpxLink that retrieves the configuration
    from a passed fakts context.

    """

    fakts_group: str
    """The fakts group within the fakts context to use for configuration"""
    fakts: Fakts
    """ The fakts context to use for configuration"""

    def configure(self, fakt: FaltsHttpXConfig) -> None:
        """Configure the link with the given fakt"""
        self.endpoint_url = fakt.endpoint_url

    async def aconnect(self, operation: Operation) -> None:
        """Connects the link to the server

        This method will retrieve the configuration from the fakts context,
        and configure the link with it. Before connecting, it will check if the
        configuration has changed, and if so, it will reconfigure the link.
        """
        fakt = await self.fakts.aget(self.fakts_group)
        assert isinstance(fakt, dict), "FaktsAIOHttpLink: fakts group is not a dict"
        self.configure(FaltsHttpXConfig(**fakt))  # type: ignore

        return await super().aconnect(operation)



from pydantic import BaseModel

from fakts_next.models import ActiveFakts


class HardFaktsGrant(BaseModel):
    """Hardcoded Fakts Grant"""

    fakts: ActiveFakts

    async def aload(self) -> ActiveFakts:
        """Loads the configuration from the hardcoded fakts"""
        return self.fakts



from pydantic import BaseModel, Field, ConfigDict, field_validator
from typing import List, Optional
import json
from hashlib import sha256


class Alias(BaseModel):
    """An alias is a way of contacting a service instance in Fakts.

    It contains the host, port, ssl flag, path and challenge.
    """

    host: str
    port: Optional[int] = None
    """The port is optional, if not set, the default port for the service will be"""
    ssl: bool = False
    """The ssl flag indicates if the service should be accessed via SSL or not. If set to True, the service will be accessed via HTTPS, otherwise it will be accessed via HTTP."""
    path: Optional[str] = None
    """The path is optional, if not set, the default path for the service will be used."""
    challenge: str = Field(
        default="",
        description="""The challenge is a string that is used to verify the alias. It should be """,
    )

    @property
    def challenge_path(self) -> str:
        """The challenge_path of the alias. Its a reachable http path that can be used to verify if the alias is accessible by the client."""
        return self.to_http_path(self.challenge)

    def to_http_path(self, append: Optional[str] = None) -> str:
        """Convert the alias to a HTTP path

        This method converts the alias to a HTTP path, which can be used to access the service.
        If the port is not set, the default port for the service will be used.
        If the ssl flag is set, the service will be accessed via HTTPS, otherwise it will be accessed via HTTP.

        Args:
            append (Optional[str], optional): An optional string to append to the path. Defaults to None.

        Returns:
            str: The HTTP path for the service
        """
        protocol = "https" if self.ssl else "http"

        url = f"{protocol}://{self.host}"
        if self.port:
            url += f":{self.port}"
        if self.path:
            url += f"/{self.path.lstrip('/')}"
        if append:
            url += f"/{append.lstrip('/')}"

        return url

    def to_ws_path(self, append: Optional[str] = None) -> str:
        """Convert the alias to a WebSocket path

        This method converts the alias to a WebSocket path, which can be used to access the service.
        If the port is not set, the default port for the service will be used.
        If the ssl flag is set, the service will be accessed via wss, otherwise it will be accessed via ws.

        Args:
            append (Optional[str], optional): An optional string to append to the path. Defaults to None.

        Returns:
            str: The WebSocket path for the service
        """
        protocol = "wss" if self.ssl else "ws"

        url = f"{protocol}://{self.host}"
        if self.port:
            url += f":{self.port}"
        if self.path:
            url += f"/{self.path.lstrip('/')}"
        if append:
            url += f"/{append.lstrip('/')}"

        return url


class Instance(BaseModel):
    """Configuration for a service in Fakts."""

    service: str
    identifier: str
    aliases: list[Alias] = []


class AuthFakt(BaseModel):
    """AuthFakt is a special kind of Fakt that is used to authenticate the user with"""

    client_id: str
    client_secret: str
    token_url: str
    scopes: List[str] = Field(default_factory=lambda: ["openid", "profile", "email"])
    """Scopes that this Fakt should request from the user"""


class SelfFakt(BaseModel):
    """SelfFakt is a special kind of Fakt that is used to identify the Fakts server itself"""

    deployment_name: str


class ActiveFakts(BaseModel):
    """The active Fakts are the Fakts that are currently active for this client"""

    self: SelfFakt
    """SelfFakt is a special kind of Fakt that is used to identify the Fakts server itself"""
    auth: AuthFakt
    instances: dict[str, Instance] = {}


class Requirement(BaseModel):
    """A requirement is a way to specify a requirement for a service instance in Fakts."""

    key: str
    service: str
    """ The service is the service that will be used to fill the key, it will be used to find the correct instance. It needs to fullfill
    the reverse domain naming scheme"""
    optional: bool = False
    """ The optional flag indicates if the requirement is optional or not. Users should be able to use the client even if the requirement is not met. """
    description: Optional[str] = None
    """ The description is a human readable description of the requirement. Will be show to the user when asking for the requirement."""


class Manifest(BaseModel):
    """A manifest for an app that can be installed in ArkitektNext

    Manifests are used to describe apps that can be installed in ArkitektNext.
    They provide information about the app, such as the
    its globally unique identifier, the version, the scopes it needs, etc.

    This Manifest is send to the Fakts server on initial app configuration,
    and is used to register the app with the Fakts server, which in turn
    will prompt the user to grant the app access to establish itself as
    an ArkitektNext app (and therefore as an OAuth2 client) (see more in the
    Fakts documentation).

    """

    version: str
    """ The version of the app TODO: Should this be a semver? """
    identifier: str
    """ The globally unique identifier of the app: TODO: Should we check for a reverse domain name? """
    scopes: List[str]
    """ Scopes that this app should request from the user """
    logo: Optional[str] = None
    """ A URL to the logo of the app TODO: We should enforce this to be a http URL as local paths won't work """
    requirements: Optional[List[Requirement]] = Field(default_factory=list)
    """ Requirements that this app has TODO: What are the requirements? """
    node_id: Optional[str] = None
    """ The node ID of the app instance, will be set automatically to the current node ID """

    description: Optional[str] = None
    """ A human readable description of the app """

    model_config = ConfigDict(extra="forbid")
    """ Configuration for the pydantic model to forbid extra fields """

    def hash(self) -> str:
        """Hash the manifest

        A manifest describes all the  metadata of an app. This method
        hashes the manifest to create a unique hash for the current configuration of the app.
        This hash can be used to check if the app has changed since the last time it was run,
        and can be used to invalidate caches.

        Returns:
            str: The hash of the manifest

        """

        unsorted_dict = self.model_dump()

        # sort the requirements
        unsorted_dict["requirements"] = sorted(
            unsorted_dict["requirements"], key=lambda x: x["key"]
        )
        # sort the scopes
        unsorted_dict["scopes"] = sorted(unsorted_dict["scopes"])

        # JSON encode the dictionary
        json_dd = json.dumps(unsorted_dict, sort_keys=True)
        # Hash the JSON encoded dictionary
        return sha256(json_dd.encode()).hexdigest()

    @field_validator("identifier", mode="after")
    def check_identifier(cls, v: str) -> str:
        """Check the identifier of the manifest
        This method checks the identifier of the manifest to ensure that it is a valid identifier.
        """
        assert "/" not in v, "The identifier should not contain a /"
        assert len(v) > 0, "The identifier should not be empty"
        assert len(v) < 256, "The identifier should not be longer than 256 characters"
        return v



import os
from typing import Optional
import pydantic
import datetime
import logging
import json
from fakts_next.models import ActiveFakts

logger = logging.getLogger(__name__)


class CacheFile(pydantic.BaseModel):
    """Cache file model"""

    fakts: ActiveFakts
    created: datetime.datetime
    hash: str = ""


class FileCache(pydantic.BaseModel):
    """Grant that caches the result of another grant

    This grant will cache the result of another grant in a file.
    It will load the grant on the first call, and then will load
    the cached version of the grant.

    Only if the cache is expired, or a "hash" value that is passed
    to the grant is different from the one in the cache, will it
    load the grant again.

    You can set the expires_in parameter to set the time in seconds
    for the cache to expire.

    FaktsRequest context parameters:
        - allow_cache: bool - whether to allow the grant to use the cache


    Attributes
    ----------
    grant : FaktsGrant
        The grant to cache
    cache_file : str
        The path to the cache file
    hash : str
        The hash to validate the cache against
    expires_in : Optional[int]
        The time in seconds for the cache to expire


    """

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)
    """The grant to cache"""

    cache_file: str = ".fakts_cache.json"
    """The path to the cache file"""
    hash: str = pydantic.Field(
        default_factory=lambda: "",
        description="Validating against the hash of the config",
    )
    """The hash to validate the cache against (if this value differes from the one in the cache, the grant will be reloaded)"""

    expires_in: Optional[int] = None
    """When should the cache expire"""

    async def aload(self) -> Optional[ActiveFakts]:
        """Loads the configuration from the grant

        It will try to load the configuration from the cache file.
        If the cache is expired, or the hash value is different from
        the one in the cache, it will load the grant again.

        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the configuration.

        Returns
        -------
        dict
            The configuration loaded from the grant.


        """

        cache = None

        if os.path.exists(self.cache_file):
            with open(self.cache_file, "r") as f:
                x = json.load(f)
                try:
                    cache = CacheFile(**x)

                    if self.hash and cache.hash != self.hash:
                        cache = None

                    elif self.expires_in:
                        if cache.created + datetime.timedelta(seconds=self.expires_in) < datetime.datetime.now():
                            cache = None

                except pydantic.ValidationError as e:
                    logger.error(f"Could not load cache file: {e}. Ignoring it")

                if cache is None:
                    return None

                return cache.fakts

    async def aset(self, value: ActiveFakts):
        """Refreshes the configuration from the grant

        This function is used to refresh the configuration from the grant.
        This is used to refresh the configuration from the grant, and should
        be used to refresh the configuration from the grant.

        The request object is used to pass information
        """

        cache = CacheFile(fakts=value, created=datetime.datetime.now(), hash=self.hash)

        with open(self.cache_file, "w+") as f:
            json.dump(json.loads(cache.model_dump_json()), f)

    async def areset(self):
        """Resets the cache

        This function is used to reset the cache.
        This is used to reset the cache, and should
        be used to reset the cache.

        The request object is used to pass information
        """

        if os.path.exists(self.cache_file):
            os.remove(self.cache_file)
        return None



import asyncio
from http import HTTPStatus
from urllib.parse import urlencode
import webbrowser
import aiohttp
import time
from pydantic import BaseModel, ConfigDict, Field, model_validator
from fakts_next.grants.remote import FaktsEndpoint
from fakts_next.grants.remote.errors import DemandError

import ssl
import certifi
from typing import Awaitable, Callable, List
from enum import Enum
from .utils import (
    acheck_supported_layers,
    print_device_code_prompt,
    print_succesfull_login,
)


DeviceCodeHook = Callable[["FaktsEndpoint", str], Awaitable[None]]
GrantedHook = Callable[["FaktsEndpoint", str], Awaitable[None]]


async def display_in_terminal(endpoint: "FaktsEndpoint", code: str) -> None:
    """A default hook that does nothing"""
    querystring = urlencode(
        {
            "device_code": code,
            "grant": "device_code",
        }
    )

    webbrowser.open_new(endpoint.base_url + "configure/?" + querystring)

    print_device_code_prompt(
        endpoint.base_url + "configure/?" + querystring,
        endpoint.base_url + "device",
        code,
    )


async def granted_in_terminal(endpoint: "FaktsEndpoint", token: str) -> None:
    """A default hook that is called when the device code is granted"""
    print_succesfull_login()


class DeviceCodeError(DemandError):
    """A base class for all device code errors"""

    pass


class DeviceCodeTimeoutError(DeviceCodeError):
    """An error that is raised when the timeout for the device code grant is reached"""

    pass


class ClientKind(str, Enum):
    """The kind of client that you want to request"""

    DEVELOPMENT = "development"
    """Tries to set up a development client (client belongs to user)"""
    WEBSITE = "website"
    """Tries to set up a website client (allows for the client to be used by anyone)"""
    DESKTOP = "desktop"


class DeviceCodeDemander(BaseModel):
    """Device Code Grant

    The device code grant is a remote grant that is able to newly establish an application
    on the fakts_next server server that support the device code grant.

    When setting up the device code grant, the user will be prompted to visit a URL and enter a code.
    If open_browser is set to True, the URL will be opened in the default browser, and automatically
    entered. Otherwise the user will be prompted to enter the code manually.

    The device code grant will then poll the fakts_next server for the status of the code. If the code is
    still pending, the grant will wait for a second and then poll again. If the code is granted, the
    token will be returned. If the code is denied, an exception will be raised.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    device_code_hook: DeviceCodeHook = Field(
        default=display_in_terminal,
        examples=True,
        description="A callback function that is called when the device code is retrieved",
    )
    granted_hook: GrantedHook = Field(
        default=granted_in_terminal,
        examples=True,
        description="A callback function that is called when the device code is granted",
    )

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    manifest: BaseModel
    """ An ssl context to use for the connection to the endpoint"""
    expiration_time_seconds: int = Field(default=300, description="The expiration time of the token in seconds")
    """The expiration time of the token in seconds"""
    redirect_uris: List[str] = Field(
        default=[],
        description="The redirect uri to use for the client if it is a desktop application",
    )
    """The redirect uri to use for the client if it is a desktop application"""
    requested_client_kind: ClientKind = Field(
        ClientKind.DEVELOPMENT,
        description="The kind of client that you want to request",
    )
    """The kind of client that you want to request. Check the ClientKind enum for more information"""

    timeout: int = 60
    """The timeout for the device code grant in seconds. If the timeout is reached, the grant will fail."""

    open_browser: bool = True
    """If set to True, the URL will be opened in the default browser (if exists). Otherwise the user will be prompted to enter the code manually."""

    @model_validator(mode="after")
    def check_requested_matches_redirect_uris(
        self: "DeviceCodeDemander",
    ) -> "DeviceCodeDemander":  # type: ignore
        """Validates and checks that either a schema_dsl or schema_glob is provided, or that allow_introspection is set to True"""
        if not self.redirect_uris and self.requested_client_kind == ClientKind.WEBSITE:
            raise ValueError("You must provide a redirect uri if you want to request a website client")

        return self

    async def arequest_code(self, endpoint: FaktsEndpoint) -> str:
        """Requests a new code from the fakts_next server.

        This method will request a new code from the fakts_next server. This code will be used to
        authenticate the user. The user will be prompted to visit a URL and enter the code.

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to fetch the token for

        Returns
        -------
        str
            The devide-code that was requested
        """

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl_context)) as session:
            while True:
                async with session.post(
                    f"{endpoint.base_url}start/",
                    json={
                        "manifest": self.manifest.model_dump(),
                        "expiration_time_seconds": self.expiration_time_seconds,
                        "redirect_uris": self.redirect_uris,
                        "requested_client_kind": self.requested_client_kind,
                        "supported_layers": await acheck_supported_layers(endpoint),
                    },
                ) as response:
                    if response.status == HTTPStatus.OK:
                        result = await response.json()
                        if result["status"] == "granted":
                            return result["code"]

                        else:
                            raise DeviceCodeError(f"Error! Could not retrieve code: {result.get('error', 'Unknown Error')}")

                    else:
                        raise DeviceCodeError(f"Server Error! Could not retrieve code {await response.text()}")

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        """Requests a token from the fakts_next server

        This method will request a token from the fakts_next server, using the device code grant.
        In the process, this grant will ask the fakts_next server to create a unique
        device code, it will then ask the user to visit a URL and enter the code.

        If open_browser is set to True, the URL will be opened in the default browser, and automatically
        entered. Otherwise the user will be prompted to enter the code manually.

        The device code grant will then poll the fakts_next server for the status of the code. If the code is
        still pending, the grant will wait for a second and then poll again. If the code is granted, the
        token will be returned. If the code is denied, an exception will be raised.

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to fetch the token for
        request : FaktsRequest
            The request to use for the fetching of the token

        Returns
        -------
        str


        """

        code = await self.arequest_code(endpoint)

        await self.device_code_hook(endpoint, code)

        start_time = time.time()

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl_context)) as session:
            while True:
                async with session.post(f"{endpoint.base_url}challenge/", json={"code": code}) as response:
                    if response.status == HTTPStatus.OK:
                        result = await response.json()
                        if result["status"] == "waiting":
                            if time.time() - start_time > self.timeout:
                                raise DeviceCodeTimeoutError("Timeout for device code grant reached.")

                            await asyncio.sleep(1)
                            continue

                        if result["status"] == "pending":
                            if time.time() - start_time > self.timeout:
                                raise DeviceCodeTimeoutError("Timeout for device code grant reached.")
                            await asyncio.sleep(1)
                            continue

                        if result["status"] == "granted":
                            await self.granted_hook(endpoint, result["token"])
                            return result["token"]

                        if result["status"] == "error":
                            raise DeviceCodeError(f"Error! Could not retrieve code: {result.get('error', 'Unknown Error')}")

                        if result["status"] == "denied":
                            raise DeviceCodeError(f"Denied! The user Denied: {result.get('message', 'Unknown Error')}")

                    else:
                        raise DeviceCodeError(f"Error! Could not retrieve code {await response.text()}")



from fakts_next.grants.remote.discovery.advertised import (
    alisten_pure,
    ListenBinding,
    Beacon,
)
from qtpy import QtWidgets, QtCore, QtGui
import asyncio
import logging
from koil.qt import qt_to_async, QtFuture
from fakts_next.grants.remote.discovery.utils import discover_url
from pydantic import BaseModel, ConfigDict
from typing import Any, Optional
from typing import Dict, List

from pydantic import Field
import ssl
import certifi
from fakts_next.grants.remote.models import FaktsEndpoint


logger = logging.getLogger(__name__)


class SelfScanWidget(QtWidgets.QWidget):  # type: ignore
    """A widget that allows the user to scan for beacons

        It has a line edit and a button. When the button is clicked,
        the line edit is scanned for a url, and if it is a valid url,
        a beacon is emitted with the url.
    from .utils import discover_url

        No validation is done on the url, so it is up to the user to
        enter a valid url.

    """

    user_beacon_added = QtCore.Signal(Beacon)  # type: ignore
    """Signal that is emitted when a new beacon is added"""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        """Constructor for SelfScanWidget"""
        super().__init__(*args, **kwargs)  # type: ignore
        self.scanlayout = QtWidgets.QHBoxLayout()
        self.lineEdit = QtWidgets.QLineEdit()
        self.addButton = QtWidgets.QPushButton("Scan")

        self.scanlayout.addWidget(self.lineEdit)
        self.scanlayout.addWidget(self.addButton)
        self.addButton.clicked.connect(self.on_add)
        self.setLayout(self.scanlayout)  # type: ignore

    def on_add(self) -> None:
        """Called when the button is clicked"""
        host = self.lineEdit.text()
        beacon = Beacon(url=host)
        self.user_beacon_added.emit(beacon)  # type: ignore


class FaktsEndpointButton(QtWidgets.QPushButton):  # type: ignore
    """A button that represents a FaktsEndpoint

    It has a title and a subtitle, and when clicked, emits
    a signal with the endpoint
    """

    accept_clicked = QtCore.Signal(FaktsEndpoint)  # type: ignore

    def __init__(self, endpoint: FaktsEndpoint, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Constructor for FaktsEndpointButton"""
        super(FaktsEndpointButton, self).__init__(parent)  # type: ignore
        self.endpoint = endpoint
        self.hovered = False
        self.is_pressed = False
        self.clicked.connect(self.on_clicked)  # type: ignore
        self.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))  # type: ignore  # Default cursor
        self.initUI()

    def initUI(self) -> None:
        """Initialize UI elements"""
        # Set the button's minimum height and other properties if needed
        self.setMinimumHeight(50)  # type: ignore
        self.setContentsMargins(10, 10, 10, 10)  # type: ignore
        self.setToolTip("Connect to " + self.endpoint.name)  # type: ignore

    def paintEvent(self, a0: Any) -> None:
        """Paint the button with the given style, title and subtitle"""
        # Call the superclass paint event to keep the button look normal
        super(FaktsEndpointButton, self).paintEvent(a0)

        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)  # type: ignore

        # Calculate the position for the title and subtitle
        title_pos_y = int(self.rect().height() / 3)
        subtitle_pos_y = int(self.rect().height() / 3 * 2)

        # Set the font for the title
        title_font = QtGui.QFont("SansSerif", 12)  # Make the title font a bit bigger
        title_font.setBold(True)
        painter.setFont(title_font)
        painter.drawText(
            QtCore.QRect(10, 10, self.width(), title_pos_y),  # type: ignore
            QtCore.Qt.AlignLeft,  # type: ignore
            self.endpoint.name,
        )

        # Set the font for the subtitle
        subtitle_font = QtGui.QFont("SansSerif", 8)
        painter.setFont(subtitle_font)
        painter.drawText(
            QtCore.QRect(10, title_pos_y + 10, self.width(), subtitle_pos_y),  # type: ignore
            QtCore.Qt.AlignLeft,  # type: ignore
            "@" + self.endpoint.base_url,
        )

    def enterEvent(self, a0: Any) -> None:
        """Sets the cursor to pointing hand when hovering over the button"""
        self.hovered = True

        self.update()  # type: ignore # Trigger a repaint
        self.setCursor(  # type: ignore
            QtGui.QCursor(QtCore.Qt.PointingHandCursor)  # type: ignore
        )  # Change cursor to pointer

    def leaveEvent(self, a0: Any) -> None:
        """Sets the cursor to arrow when not hovering over the button"""
        self.hovered = False
        self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))  # type: ignore  # Change cursor to arrow
        self.update()  # type: ignore# Trigger a repaint# type: ignore

    def mousePressEvent(self, e: Any) -> None:
        """Sets the button to pressed when clicked"""
        self.is_pressed = True
        self.update()  ## type: ignore Trigger a repaint
        super(FaktsEndpointButton, self).mousePressEvent(e)  # type: ignore

    def mouseReleaseEvent(self, e: Any) -> None:
        """Sets the button to not pressed when released"""
        self.is_pressed = False
        self.update()  # type: ignore # Trigger a repaint
        super(FaktsEndpointButton, self).mouseReleaseEvent(e)  # type: ignore

    def on_clicked(self) -> None:
        """Called when the button is clicked"""
        self.accept_clicked.emit(self.endpoint)  # type: ignore


class SelectBeaconWidget(QtWidgets.QDialog):  # type: ignore
    """A widget that allows the user to select a beacon

    It has a list of buttons, each representing a beacon.
    When a button is clicked, a signal is emitted with the beacon
    """

    new_advertised_endpoint = QtCore.Signal(FaktsEndpoint)  # type: ignore
    new_local_endpoint = QtCore.Signal(FaktsEndpoint)  # type: ignore

    def __init__(
        self,
        *args,  # type: ignore
        settings: Optional[QtCore.QSettings] = None,  # type: ignore
        **kwargs,  # type: ignore
    ) -> None:
        """Constructor for SelectBeaconWidget"""
        super().__init__(*args, **kwargs)  # type: ignore
        self.setWindowTitle("Search Endpoints...")
        self.hide_coro = qt_to_async(self.hide_callback)
        self.show_error_coro = qt_to_async(self.show_error)
        self.clear_endpoints_coro = qt_to_async(self.clear_endpoints)
        self.select_endpoint = qt_to_async(self.demand_selection_of_endpoint)  # type: ignore
        self.settings = settings  # type: ignore

        self.select_endpoint_future = None

        self.new_advertised_endpoint.connect(self.on_new_endpoint)  # type: ignore
        self.new_local_endpoint.connect(self.on_new_endpoint)  # type: ignore

        self.endpoints: List[FaktsEndpoint] = []

        self.wlayout = QtWidgets.QVBoxLayout()

        self.endpointLayout = QtWidgets.QVBoxLayout()

        self.scanWidget = SelfScanWidget()

        QBtn = QtWidgets.QDialogButtonBox.Cancel  # type: ignore
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)  # type: ignore
        self.buttonBox.rejected.connect(self.on_reject)

        endgroup = QtWidgets.QGroupBox("Select")
        endgroup.setLayout(self.endpointLayout)
        endgroup.setStyleSheet("padding: 10px;")

        self.scan_layout = QtWidgets.QVBoxLayout()
        self.scan_layout.addWidget(self.scanWidget)

        scangroup = QtWidgets.QGroupBox("Or Scan")
        scangroup.setLayout(self.scan_layout)

        self.wlayout.addWidget(endgroup)
        self.wlayout.addWidget(scangroup)
        self.wlayout.addWidget(self.buttonBox)
        self.setLayout(self.wlayout)  # type: ignore

    def hide_callback(self, future: QtFuture[bool]) -> None:  # type: ignore
        self.hide()
        future.resolve(True)  # type: ignore

    def clearLayout(self, layout: QtWidgets.QVBoxLayout) -> None:
        """Clear the layout"""
        while layout.count():
            child = layout.takeAt(0)
            if child is None:
                continue

            widget = child.widget()
            if widget is not None:
                widget.deleteLater()

    def clear_endpoints(self, future: QtFuture[bool]) -> None:
        """Clear the endpoints"""
        self.clearLayout(self.endpointLayout)
        self.endpoints = []
        future.resolve(True)  # type: ignore

    def show_me(self) -> None:
        """A function that shows the widget"""
        self.show()  # type: ignore

    def show_error(self, future: QtFuture[None], error: Exception) -> None:
        """Show an error message

        Parameters
        ----------
        error : Exception
            Display the error message
        """
        self.show()  # type: ignore
        QtWidgets.QMessageBox.critical(self, "Error", str(error))
        future.resolve(None)  # type: ignore

    def demand_selection_of_endpoint(self, future: QtFuture) -> None:  # type: ignore
        """Is called when the user should select an endpoint. I.e. when the demander
        is called programmatically"""
        self.select_endpoint_future = future  # type: ignore
        self.show()  # type: ignore# type: ignore

    def on_endpoint_clicked(self, item: FaktsEndpoint) -> None:
        """Called when an endpoint button is clicked

        Will resolve the future with the endpoint



        """
        if self.select_endpoint_future:  # type: ignore
            self.select_endpoint_future.resolve(item)  # type: ignore

    def on_reject(self) -> None:
        """Called when the user rejects the dialog

        Will reject the future"""
        if self.select_endpoint_future:  # type: ignore
            self.select_endpoint_future.reject(  # type: ignore
                Exception("User cancelled the this Grant without selecting a Beacon")
            )  # type: ignore
        self.reject()  # type: ignore

    def closeEvent(self, a0: Any) -> None:
        """Called when the window is closed. Will automatically reject the future"""
        if self.select_endpoint_future:  # type: ignore
            self.select_endpoint_future.reject(  # type: ignore
                Exception("User cancelled the this Grant without selecting a Beacon")
            )  # type: ignore

        a0.accept()  # let the window close

    def on_new_endpoint(self, config: FaktsEndpoint) -> None:
        """A callback that is called when a new endpoint is discovered

        Creates a button for the endpoint and adds it to the layout


        """
        self.clearLayout(self.endpointLayout)

        self.endpoints.append(config)

        for endpoint in self.endpoints:
            widget = FaktsEndpointButton(endpoint)

            self.endpointLayout.addWidget(widget)
            widget.accept_clicked.connect(self.on_endpoint_clicked)  # type: ignore


async def wait_first(*tasks) -> Any:  # type: ignore
    """Return the result of first async task to complete with a non-null result"""
    # Get first completed task(s)
    done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)  # type: ignore

    # Cancel pending tasks
    for task in pending:  # type: ignore
        task.cancel()  # type: ignore

    # Wait for pending tasks to be cancelled
    await asyncio.gather(*pending, return_exceptions=True)  # type: ignore

    # Return first completed task result
    for task in done:  # type: ignore
        return task.result()  # type: ignore


class QtSelectableDiscovery(BaseModel):
    """A QT based discovery that will cause the user to select an endpoint
    from a list of network discovered endpoints, as well as endpoints
    that the user manually enters.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    binding: ListenBinding = Field(default_factory=ListenBinding)
    """The address to bind to"""
    strict: bool = False
    """Should we error on bad Beacons"""
    discovered_endpoints: Dict[str, FaktsEndpoint] = Field(default_factory=dict)
    """A cache of discovered endpoints"""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )
    additional_beacons: List[str] = Field(default_factory=lambda: ["localhost:11000", "localhost:11001", "localhost:8000"])
    widget: SelectBeaconWidget

    async def emit_endpoints(self) -> None:
        """A long running task that will emit endpoints that are discovered
        through the network and emit them to be displayed in the widget.
        """
        if self.additional_beacons:
            for i in self.additional_beacons:
                try:
                    endpoint = await discover_url(
                        i,
                        self.ssl_context,
                        auto_protocols=self.auto_protocols,
                        allow_appending_slash=self.allow_appending_slash,
                        timeout=self.timeout,
                    )
                    self.widget.new_local_endpoint.emit(endpoint)  # type: ignore
                except Exception as e:
                    logger.info(f"Could not connect to localhost: {e}")

        try:
            try:
                binding = self.binding
                async for beacon in alisten_pure(binding, strict=self.strict):
                    try:
                        if beacon.url in self.additional_beacons:
                            # we already did this one
                            continue
                        endpoint = await discover_url(
                            beacon.url,
                            self.ssl_context,
                            auto_protocols=self.auto_protocols,
                            allow_appending_slash=self.allow_appending_slash,
                            timeout=self.timeout,
                        )
                        self.widget.new_advertised_endpoint.emit(endpoint)  # type: ignore
                    except Exception as e:
                        logger.info(f"Could not connect to beacon: {beacon.url} {e}")
            except Exception:
                logger.exception("Error in discovery")
                return None

        except Exception as e:
            logger.exception(e)
            raise e

    async def await_user_definition(
        self,
    ) -> Optional[FaktsEndpoint]:
        """Await the user to define a beacon. This will cause the widget to listen for
        user input.

        Returns
        -------
        FaktsEndpoint
            The endpoint that the user defined

        """

        async for beacon in self.widget.beacon_user.aiterate():  # type: ignore
            try:
                return await discover_url(
                    beacon.url,  # type: ignore
                    self.ssl_context,
                    auto_protocols=self.auto_protocols,
                    allow_appending_slash=self.allow_appending_slash,
                    timeout=self.timeout,
                )
            except Exception as e:
                await self.widget.show_error_coro.acall(e)  # type: ignore
                logger.error(f"Could not connect to beacon: {beacon.url} {e}")  # type: ignore
                continue

        return None

    async def adiscover(self) -> FaktsEndpoint:
        """Discover the endpoint

        This discovery will cause the attached widget to be shown, and
        then the user is able to select and endpoint that is discovered
        automatically or by the user.

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the discovery process (is not used)

        Returns
        -------
        FaktsEndpoint
            A valid endpoint
        """

        emitting_task = asyncio.create_task(self.emit_endpoints())
        try:
            await self.widget.clear_endpoints_coro.acall()

            try:
                select_endpoint_task = asyncio.create_task(  # type: ignore
                    self.widget.select_endpoint.acall()  # type: ignore
                )
                user_definition_task = asyncio.create_task(self.await_user_definition())

                endpoint: FaktsEndpoint = await wait_first(select_endpoint_task, user_definition_task)

                await self.widget.hide_coro.acall()  # type: ignore

            finally:
                emitting_task.cancel()
                try:
                    await emitting_task
                except asyncio.CancelledError:
                    logger.info("Cancelled the Discovery task")

            return endpoint
        except Exception as e:
            logger.exception(e)
            emitting_task.cancel()

            try:
                await emitting_task
            except asyncio.CancelledError:
                logger.info("Cancelled the Discovery task")

            raise e



from pydantic import ConfigDict, Field
import ssl
import certifi
import aiohttp
from fakts_next.grants.remote.errors import ClaimError
from fakts_next.grants.remote.models import FaktsEndpoint
from pydantic import BaseModel

from fakts_next.models import ActiveFakts


class ClaimEndpointClaimer(BaseModel):
    """A claimer that claims the configuration from the endpoint

    This claimer is used to claim the configuration from the endpoint.
    This is the default claimer, and it is used by the default
    Remote Grants.


    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    async def aclaim(
        self,
        token: str,
        endpoint: FaktsEndpoint,
    ) -> ActiveFakts:
        """Claims the configuration from the endpoint

        Parameters
        ----------
        token : str
            The token to use to claim the configuration
        endpoint : FaktsEndpoint
            The endpoint to claim the configuration from
        request : FaktsRequest
            The request to use to claim the configuration

        Returns
        -------
        Dict[str, FaktValue]
            The configuration

        Raises
        ------
        ClaimError
            An error occured while claiming the configuration
        """

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl_context)) as session:
            async with session.post(
                f"{endpoint.base_url}claim/",
                json={
                    "token": token,
                    "secure": endpoint.base_url.startswith("https"),
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if "status" not in data:
                        raise ClaimError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise ClaimError(data["message"])
                    if status == "granted":
                        return ActiveFakts(**data["config"])
                    if status == "denied":
                        raise ClaimError("Access denied")

                    raise ClaimError(f"Unexpected status: {status}")
                else:
                    raise ClaimError("Error! Coud not claim this app on this endpoint")



"""Remote Discovery mechanisms

This module contains the different discovery mechanisms
that can be used to discover the endpoint to connect to.

These are NOT grants, and are used by the remote grants
to discover the endpoint to connect to.

"""

from .static import StaticDiscovery
from .well_known import WellKnownDiscovery
from .advertised import FirstAdvertisedDiscovery

__all__ = ["StaticDiscovery", "FirstAdvertisedDiscovery", "WellKnownDiscovery"]



import pydantic
from typing import Dict, Any
import datetime


class CacheModel(pydantic.BaseModel):
    """Cache file model"""

    config: Dict[str, Any]
    created: datetime.datetime
    hash: str = ""



"""Fakts package.

Fakts is a configuration management library. It allows you to
load configuration from different sources and use it in your
application.

In comparison to other configuration management libraries, Fakts
is designed to be used in an async environment, or with primarily
async grants (e.g. a database). It also allows you to use multiple
grants at the same time, and will merge the results together.

Fakts comes also with a few remote grants, that allow you to connect
to a remote configuration server, and fetch the configuration from
there, and follows a similar design the oauth2 protocol, to allow for
safe and secure configuration management.

"""

from .fakts import Fakts, FaktsGrant, get_current_fakts_next
from .errors import FaktsError
from .grants import EnvGrant, GrantError


__all__ = [
    "Fakts",
    "FaktsGrant",
    "EnvGrant",
    "GrantError",
    "get_current_fakts_next",
    "FaktsError",
]



from pydantic import BaseModel
from fakts_next.fakts import Fakts
from rath.links.subscription_transport_ws import SubscriptionTransportWsLink
from rath.operation import Operation


class WebsocketHttpConfig(BaseModel):
    """WebsocketHttpConfig"""

    ws_endpoint_url: str


class FaktsWebsocketLink(SubscriptionTransportWsLink):
    """FaktsWebsocketLink


    A FaktsWebsocketLink is a SubscriptionTransportWsLink that retrieves the configuration
    from a passed fakts context.

    """

    fakts: Fakts
    """The fakts context to use for configuration"""
    fakts_group: str = "websocket"
    """ The fakts group within the fakts context to use for configuration """

    def configure(self, fakt: WebsocketHttpConfig) -> None:
        """Configure the link with the given fakt"""
        self.ws_endpoint_url = fakt.ws_endpoint_url

    async def aconnect(self, operation: Operation) -> None:
        """Connects the link to the server

        This method will retrieve the configuration from the fakts context,
        and configure the link with it. Before connecting, it will check if the
        configuration has changed, and if so, it will reconfigure the link.
        """
        fakt = await self.fakts.aget(self.fakts_group)
        assert isinstance(fakt, dict), "FaktsAIOHttpLink: fakts group is not a dict"
        self.configure(WebsocketHttpConfig(**fakt))  # type: ignore

        return await super().aconnect(operation)



from fakts_next import Fakts
from rath.links.auth import AuthTokenLink
from rath.operation import Operation


class FaktsAuthLink(AuthTokenLink):
    """faktsAuthLink is a link that retrieves a token from oauth2 and sends it to the next link."""

    fakts: Fakts

    async def aload_token(self, operation: Operation) -> str:
        """Retrieves the token from herre"""
        fakts = self.fakts
        return await fakts.aget_token()

    async def arefresh_token(self, operation: Operation) -> str:
        """Refreshes the token from herre"""
        fakts = self.fakts
        return await fakts.arefresh_token()



from typing import Dict
from fakts_next.protocols import FaktValue
from pydantic import BaseModel


class BaseFaktsGrant(BaseModel):
    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the grant

        Depending on the grant, this function may load the configuration
        from a file, from a remote endpoint, from a database, etc, the
        implementation of the grant will determine how the configuration
        is loaded.

        Generally, the grant should use preconfigured values to set the
        configuration retrievel logic, and should not use the request
        object to determine how to load the configuration.

        The request object is used to pass information between different
        grants, and should only be used to forward conditional information
        like "skip cache" or "force refresh". Which are mainly handled
        by meta grants.



        Returns
        -------
        dict
            The configuration loaded from the grant.

        Raises
        ------

        GrantError
            If the grant failed to load the configuration.+



        """
        raise NotImplementedError("aload not implemented by the grant")

    async def arefresh(self) -> Dict[str, FaktValue]:
        """Refreshes the configuration from the grant

        This function is used to refresh the configuration from the grant.
        This is used to refresh the configuration from the grant, and should
        be used to refresh the configuration from the grant.

        The request object is used to pass information
        """
        return await self.aload()



import ssl
import certifi
from pydantic import ConfigDict, Field, BaseModel
import logging
from typing import List
from .utils import discover_url
from fakts_next.grants.remote import FaktsEndpoint

logger = logging.getLogger(__name__)


class WellKnownDiscovery(BaseModel):
    """A discovery that uses the well-known endpoint

    A well-known endpoint is a special endpoint that is used to discover
    the actual endpoint. This is useful for example in scenarious where the
    fakts_next server might change its location, and the client needs to find it,
    easily.

    This discovery mechanism also provides ways to configure the discovery
    process, such as the timeout, the ssl context, and the protocols to try
    if no protocol is specified in the url.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    url: str
    """The url of the well-known endpoint"""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    """If the url does not end with a slash, should we append one? A well-known endpoint should end with a slash"""
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    """ If no protocol is specified, we will try to connect to the following protocols"""
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )
    """A timeout for the connection to the well-known endpoint. Applies to each protocol"""

    async def adiscover(self) -> FaktsEndpoint:
        """Discover the endpoint

        This method will try to discover the endpoint using the well-known
        endpoint. It will try to connect to the well-known endpoint using
        the protocols specified in the configuration, and the timeout
        specified in the configuration.

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the discovery process (is not used)

        Returns
        -------
        FaktsEndpoint
            A valid endpoint
        """

        return await discover_url(
            self.url,
            self.ssl_context,
            auto_protocols=self.auto_protocols,
            allow_appending_slash=self.allow_appending_slash,
            timeout=self.timeout,
        )



from fakts_next.models import ActiveFakts
from ..errors import GrantError
import logging

from .models import Demander, Discovery, Claimer
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)


Token = str
EndpointUrl = str


class RemoteGrantError(GrantError):
    """Base class for all remotegrant errors"""


class RemoteGrant(BaseModel):
    """Abstract base class for remote grants

    A Remote grant is a grant that connects to a fakts_next server,
    and tires to establishes a secure relationship with it.

    This is a highly configurable grant, that can be used to
    dynaimcially *discover* the endpoint, *demand* a token
    to access a token, and then *claim* the configuration
    from the endpoint.

    This grant is highly configurable, and can be used to
    implement any kind of remote grant.

    You can use a specific builder to build a remote grant
    that fits your needs.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    discovery: Discovery
    """The discovery mechanism to use for finding the endpoint"""

    demander: Demander
    """The demander mechanism to use for demanding the token FROM the endpoint"""

    claimer: Claimer
    """The claimer mechanism to use for claiming the token FROM the endpoint"""

    async def aload(self) -> ActiveFakts:
        """Load the configuration

        This function will first discover the endpoint, then demand a token from it,
        and then claim the configuration from it.

        Parameters
        ----------
        request : FaktsRequest
            The request to use for the load

        Returns
        -------
        Dict[str, FaktValue]
            The configuration that was claimed from the endpoint



        """
        try:
            endpoint = await self.discovery.adiscover()
        except Exception as e:
            raise RemoteGrantError(f"Could not discover endpoint: {e}") from e

        token = await self.demander.ademand(endpoint)

        return await self.claimer.aclaim(token, endpoint)



from fakts_next.grants.errors import GrantError
import os
import logging
from fakts_next.protocols import FaktValue
from typing import Dict, Any, List
from pydantic import BaseModel

logger = logging.getLogger(__name__)


def nested_set(dic: Dict[str, Any], keys: List[str], value: Any) -> None:
    """Sets a value in a nested dictionary (inplace version)

    Parameters
    ----------
    dic : Dict
        The dictionary to set the value in
    keys : List[str]
        The keys to set the value in
    value : Any
        The value to set
    """
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value
    return None


class EnvGrant(BaseModel):
    """Extras a configuration tree from the current environment
    variables.

    It will load all the environment variables that start with the `prepend` string,
    and will parse them as a configuration tree (splitting the key by the `delimiter`)
    it will also convert the keys to lowercase.

     E.g.

        ```env
        FAKTS_GROUP_NAME__KEY_NAME=value
        FAKTS_OTHER__KEY_NAME2=value2
        ```
        ```python

        grant = EnvGrant(delimiter__="__", prepend="FAKTS_") # the default values

        with Fakts(grant=grants) as fakts_next:
            config = grant.get() # accessing the config
            print(config["group_name"]["key_name"]) # value

            print(grant.get("other.key_name2") # value2
        ```
        ```

    """

    prepend: str = "FAKTS_"
    delimiter: str = "__"

    async def aload(self) -> Dict[str, FaktValue]:
        """Loads the configuration from the environment variables.

        It will load all the environment variables that start with the `prepend` string,
        and will parse them as a configuration tree (splitting the key by the `delimiter`)

        E.g.

        ```env
        FAKTS_GROUP_NAME__KEY_NAME=value
        FAKTS_OTHER__KEY_NAME2=value2
        ```
        ```python

        grant = EnvGrant(delimiter__="__", prepend="FAKTS_") # the default values

        with Fakts(grant=grants) as fakts_next:
            config = grant.get() # accessing the config
            print(config["group_name"]["key_name"]) # value

            print(grant.get("other.key_name2") # value2
        ```

        Returns
        -------
        Dict[str, FaktValue]
            The loaded fakts_next

        Raises
        ------
        GrantError
            If the environment variables could not be loaded, because
            of a nested key that is not a dictionary.
        """
        try:
            data: Dict[str, FaktValue] = {}

            for key, value in os.environ.items():
                if self.prepend:
                    if not key.startswith(self.prepend):
                        continue
                    key = key[len(self.prepend) :]

                path = list(map(lambda x: x.lower(), key.split(self.delimiter)))

                nested_set(data, path, value)

            return data

        except Exception as e:
            raise GrantError(f"Could not load from env: {e}") from e



import logging
from qtpy import QtCore

from fakts_next.grants.remote.models import FaktsEndpoint


from typing import Dict, Optional

import json
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)


class EndpointDefaults(BaseModel):
    """A serialization helper for the
    default token store"""

    default_token: Dict[str, str] = {}


class QTSettingTokenStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts_next grant"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    settings: QtCore.QSettings  # type: ignore
    """The settings to use to store the tokens"""
    save_key: str
    """The key to use to store the tokens"""

    async def aput_default_token_for_endpoint(self, endpoint: FaktsEndpoint, token: str) -> None:
        """A function that puts the default token for an endpoint
        from the settings

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to put the token for
        token : str
            The token to put, or None to delete the token
        """

        un_storage = self.settings.value(self.save_key, None)  # type: ignore
        if not un_storage:
            storage = EndpointDefaults()
        else:
            try:
                storage = EndpointDefaults(**json.loads(un_storage))  # type: ignore
            except Exception as e:
                print("Error loading token store", e)
                storage = EndpointDefaults()

        if endpoint.base_url in storage.default_token:
            del storage.default_token[endpoint.base_url]
        else:
            storage.default_token[endpoint.base_url] = token

        self.settings.setValue(self.save_key, storage.model_dump_json())  # type: ignore

    async def aget_default_token_for_endpoint(self, endpoint: FaktsEndpoint) -> Optional[str]:
        """A function that gets the default token for an endpoint
        from the settings

        Parameters
        ----------
        endpoint : FaktsEndpoint
            The endpoint to get the token for

        Returns
        -------
        Optional[str]
            The token for the endpoint, or None if there is no token
        """

        un_storage = self.settings.value(self.save_key, None)  # type: ignore
        if not un_storage:
            return None
        try:
            storage = EndpointDefaults(**json.loads(un_storage))  # type: ignore
            if endpoint.base_url in storage.default_token:
                return storage.default_token[endpoint.base_url]
        except Exception as e:
            print(e)
            return None

        return None



from pydantic import Field
from typing import Any, Dict
from fakts_next.grants.remote.models import FaktsEndpoint, ActiveFakts
from pydantic import BaseModel


class StaticClaimer(BaseModel):
    """A claimer that always claims
    the same configuration

    This is mostly used for testing purposes.

    """

    value: ActiveFakts = Field(
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    async def aclaim(
        self,
        token: str,
        endpoint: FaktsEndpoint,
    ) -> ActiveFakts:
        """Claim the configuration from the endpoint"""

        return self.value



"""Remote Claimer mechanisms

This module contains the different cleaimer mechanisms
that can be used to retrieve a configuration from
a previously discovered endpoint and its token.


These are NOT grants, but are used by the remote grant.

"""

from .static import StaticClaimer
from .post import ClaimEndpointClaimer

__all__ = ["StaticClaimer", "ClaimEndpointClaimer"]



from fakts_next.grants.remote.models import FaktsEndpoint
import aiohttp
import logging
import ssl
from fakts_next.grants.remote.errors import DiscoveryError
from typing import Optional, List, Tuple

logger = logging.getLogger(__name__)


async def check_wellknown(url: str, ssl_context: ssl.SSLContext, timeout: int = 4) -> FaktsEndpoint:
    """Check the well-known endpoint

    This function will check the well-known endpoint and return the endpoint
    if it is valid. If it is not valid, it will raise an exception.

    Parameters
    ----------
    url : str
        Url to check
    ssl_context : ssl.SSLContext
        The ssl context to use for the connection
    timeout : int, optional
        The timeout for the connection , by default 4

    Returns
    -------
    FaktsEndpoint
        A valid endpoint

    Raises
    ------
    DiscoveryError
    """
    url = f"{url}.well-known/fakts"

    async with aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=ssl_context),
        headers={"User-Agent": "Fakts/0.1", "Accept": "application/json"},
    ) as session:
        async with session.get(
            url,
            timeout=aiohttp.ClientTimeout(total=timeout),
        ) as resp:
            if resp.status == 200:
                data = await resp.json()

                if "name" not in data:
                    logger.error(f"Malformed answer: {data}")
                    raise DiscoveryError("Malformed Answer")

                return FaktsEndpoint(**data)

            else:
                logger.error(f"Could not retrieve on the endpoint: {resp.status}")
                raise DiscoveryError(f"Error! We could not retrieve the endpoint. {url} ")


async def discover_url(
    url: str,
    ssl_context: ssl.SSLContext,
    auto_protocols: Optional[List[str]] = None,
    allow_appending_slash: bool = False,
    timeout: int = 4,
) -> FaktsEndpoint:
    """Discover the endpoint from the url

    This function will try to discover the endpoint from the url. If the url
    does not contain a protocol, it will try to use the auto protocols to
    discover the endpoint.

    Parameters
    ----------
    url : str
        The (base) url to discover
    ssl_context : ssl.SSLContext
        The ssl context to use for the connection
    auto_protocols : Optional[List[str]], optional
        The protocols to try (e.g. http https), by default None
    allow_appending_slash : bool, optional
        Should we autoappend a slash if the ur does not conain it, by default False
    timeout : int, optional
        How long to wait to consider a connection not valid, by default 4

    Returns
    -------
    FaktsEndpoint
        The endpoint

    Raises
    ------
    DiscoveryError
    """

    if "://" not in url:
        logger.info(f"No protocol specified on {url}")
        if not auto_protocols or len(auto_protocols) == 0:
            raise DiscoveryError("No protocol specified and no auto protocols specified")

        errors: list[Tuple[str, Exception]] = []

        for protocol in auto_protocols:
            logger.info(f"Trying to connect to {protocol}://{url}")
            try:
                if allow_appending_slash and not url.endswith("/"):
                    url = f"{url}/"

                return await check_wellknown(f"{protocol}://{url}", ssl_context, timeout=timeout)
            except Exception as e:
                logger.info(f"Could not connect to {protocol}://{url}")
                errors.append((protocol, e))
                continue

        errors_string = "\n".join([f"- {protocol}://{url}\n  " + str(e) for protocol, e in errors])

        raise DiscoveryError(f"Could not connect via any protocol: \n{errors_string}")

    if allow_appending_slash and not url.endswith("/"):
        url = f"{url}/"

    return await check_wellknown(url, ssl_context, timeout=timeout)



"""Provides  a fakts implementaiton of the aiohttp link"""

from typing import Any, Dict, Optional
from pydantic import BaseModel
from fakts_next.fakts import Fakts
from rath.links.aiohttp import AIOHttpLink
from rath.operation import Operation


class AioHttpConfig(BaseModel):
    """AioHttpConfig

    AioHttpConfig is a Fakt that can be used to configure the aiohttp client.
    """

    endpoint_url: str
    """The endpoint url to use for the aiohttp client"""


class FaktsAIOHttpLink(AIOHttpLink):
    """FaktsAIOHttpLink

    A FaktsAIOHttpLink is a link that retrieves the configuration
    from a sorounding fakts context.

    """

    fakts: Fakts
    """The fakts context to use for configuration"""

    fakts_group: str
    """ The fakts group within the fakts context to use for configuration """

    _old_fakt: Optional[Dict[str, Any]] = None

    async def aconfigure(self) -> None:
        """Configure the link with the given fakt"""

        alias = await self.fakts.aget_alias(self.fakts_group)
        self.endpoint_url = alias.to_http_path("graphql")

    async def aconnect(self, operation: Operation) -> None:
        """Connects the link to the server

        This method will retrieve the configuration from the fakts context,
        and configure the link with it. Before connecting, it will check if the
        configuration has changed, and if so, it will reconfigure the link.
        """

        await self.aconfigure()

        return await super().aconnect(operation)



import asyncio
import contextvars
import logging
from typing import Any, Dict, Type
from koil.composition import KoiledModel
from koil.helpers import unkoil
from pydantic import Field
from typing import Optional
from fakts_next.errors import GroupNotFound, NoFaktsFound
from fakts_next.cache.nocache import NoCache
from .protocols import FaktsCache, FaktsGrant
from .models import ActiveFakts, Alias, Manifest
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from oauthlib.oauth2.rfc6749.errors import InvalidClientError
import aiohttp
from oauthlib.common import urldecode
import ssl
import certifi
from ssl import SSLContext


logger = logging.getLogger(__name__)
current_fakts_next: contextvars.ContextVar[Optional["Fakts"]] = contextvars.ContextVar(
    "current_fakts_next", default=None
)


class Fakts(KoiledModel):
    """Fakts is any asynchronous configuration loader.

    Fakts provides a way to concurrently load and access configuration from different
    sources in async and sync environments.

    It is used to load configuration from a grant, and to access it in async
    and sync code.

    A grant constitutes the way to load configuration. It can be a local config file
    (eg. yaml, toml, json), environemnt variables, a remote configuration (eg. from
    a fakts server) a database, or any other source.  It will be loaded either on
    call to `load`,  or on  a call to `get` (if auto_load is set to true).

    Additionaly you can compose grants with the help of meta grants in order to
    load configuration from multiple sources.

    Example:
        ```python
        async with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.aget("group_name")
        ```

        or

        ```python
        with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.get("group_name")
        ```

    Fakts should be used as a context manager, and will set the current fakts context
    variable to itself, letting you access the current fakts instance from anywhere in
    your code (async or sync). To understand how the async sync code access work,
    please check out the documentation for koil.


    Example:
        ```python
        async with Fakts(grant=FailsafeGrant(
            grants=[
                EnvGrant(),
                YamlGrant("config.yaml")
            ]
        )) as fakts:
            config = await fakts.get("group_name")
        ```
        In this example fakts will load the configuration from the environment
        variables first, and if that fails, it will load it from the yaml file.


    """

    cache: FaktsCache = Field(default_factory=NoCache, exclude=True)

    """" Requirmements """
    manifest: Manifest

    """"The manifest of the fakts. This is used to describe the fakts and its capabilities."""
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )

    grant: FaktsGrant
    """The grant to load the configuration from"""

    hard_fakts: ActiveFakts | None = Field(default=None, exclude=True)
    """Hard fakts are fakts that are set by the user and cannot be overwritten by grants"""

    loaded_fakts: ActiveFakts | None = Field(default=None, exclude=True)
    """The currently loaded fakts. Please use `get` to access the fakts"""

    alias_map: Dict[str, Alias] = Field(
        default_factory=dict,
        exclude=True,
        description="Map of service names to active aliases",
    )

    loaded_token: Optional[str] = Field(
        default=None, exclude=True, description="The currently loaded token"
    )

    allow_auto_load: bool = Field(
        default=True, description="Should we autoload on get?"
    )
    """Should we autoload the grants on a call to get?"""

    load_on_enter: bool = False
    """Should we load on connect?"""
    delete_on_exit: bool = False
    """Should we delete on connect?"""

    refetch_on_group_not_found: bool = False

    _loaded: bool = False
    _lock: Optional[asyncio.Lock] = None
    _token_lock: Optional[asyncio.Lock] = None
    _alias_lock: Optional[asyncio.Lock] = None

    async def arefresh_token(self, allow_refresh: bool = True) -> str:
        """Refresh the authentication token for a service (async)"""
        """Get Authentikation Token for a service (async)"""
        assert self._lock is not None, (
            "You need to enter the context first before calling this function"
        )
        async with self._lock:
            if not self.loaded_fakts:
                try:
                    await self.aload()
                except Exception as e:
                    logger.error(e, exc_info=True)
                    raise e

        assert self.loaded_fakts, "No fakts loaded yet. Please call load() first."

        scope = " ".join(self.loaded_fakts.auth.scopes)

        auth_client = BackendApplicationClient(
            client_id=self.loaded_fakts.auth.client_id,
            scope=scope,
        )

        token_url = self.loaded_fakts.auth.token_url

        body = auth_client.prepare_request_body(
            client_secret=self.loaded_fakts.auth.client_secret,
            client_id=self.loaded_fakts.auth.client_id,
        )

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }

        data = dict(urldecode(body))

        print("Challening for token with data:", data, token_url)

        # Create an OAuth2 session for the OSF
        async with aiohttp.ClientSession(
            connector=(
                aiohttp.TCPConnector(ssl=self.ssl_context) if self.ssl_context else None
            ),
            headers=headers,
        ) as session:
            async with session.post(
                token_url,
                data=data,
                auth=aiohttp.BasicAuth(
                    self.loaded_fakts.auth.client_id,
                    self.loaded_fakts.auth.client_secret,
                ),
            ) as resp:
                text = await resp.text()

                try:
                    auth_client.parse_request_body_response(text, scope=scope)
                except InvalidClientError as e:
                    logger.error(
                        f"Invalid client error while trying to get token for {self.loaded_fakts.auth.client_id} with response: {text}. We are trying to reload the fakts."
                    )
                    if not allow_refresh:
                        raise e

                    await self.aload(reload=True)
                    return await self.arefresh_token(allow_refresh=False)

                token = auth_client.token
                self.loaded_token = str(token["access_token"])
                return str(token["access_token"])

    async def achallenge_alias(self, alias: Alias) -> bool:
        async with aiohttp.ClientSession(
            connector=(
                aiohttp.TCPConnector(ssl=self.ssl_context) if self.ssl_context else None
            ),
            headers={
                "Accept": "application/json",
                "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
            },
        ) as session:
            async with session.get(
                alias.challenge_path,
            ) as resp:
                # Check status code
                if resp.status != 200:
                    logger.error(
                        f"Failed to challenge alias {alias} with status code {resp.status}"
                    )
                    return False
                else:
                    return True

        return False

    async def aget_token(self) -> str:
        """Refresh the authentication token for a service (async)"""
        """Get Authentikation Token for a service (async)"""
        assert self._token_lock is not None, (
            "You need to enter the context first before calling this function"
        )
        async with self._token_lock:
            if not self.loaded_token:
                try:
                    await self.arefresh_token()
                except Exception as e:
                    logger.error(e, exc_info=True)
                    raise e

        assert self.loaded_token, "No token loaded yet. Please call load() first."
        return self.loaded_token

    async def aload(self, reload: bool = False) -> ActiveFakts:
        """Load the fakts from the grant (async)

        This method will load the fakts from the grant, and set the loaded_fakts
        attribute to the loaded fakts. If the fakts are already loaded, it will
        return the loaded fakts.

        Returns:
            ActiveFakts: The loaded fakts
        """
        if self.cache and not reload:
            cached_fakts = await self.cache.aload()
            if cached_fakts:
                self.loaded_fakts = cached_fakts
                return self.loaded_fakts

        self.loaded_fakts = await self.grant.aload()
        await self.cache.aset(self.loaded_fakts)
        return self.loaded_fakts

    async def arefresh(self):
        """Refresh the fakts (async)

        This method will refresh the fakts by calling the grant's refresh method.
        It will also update the loaded_fakts attribute and cache.
        """
        if not self.loaded_fakts:
            await self.aload(reload=True)
        else:
            self.loaded_fakts = await self.grant.arefresh(self.loaded_fakts)
            await self.cache.aset(self.loaded_fakts)

        return self.loaded_fakts

    async def arefresh_alias(
        self,
        service_name: Optional[str] = None,
        omit_challenge: bool = False,
        cache: bool = True,
        store: bool = True,
    ) -> Alias:
        """Refresh the alias for a service (async)

        This method will refresh the alias for a service, by calling the challenge path
        of the alias. If the challenge fails, it will raise an exception.

        Args:
            service_name (str): The name of the service to refresh the alias for
            cache (bool, optional): Should we cache the alias? Defaults to True.
            store (bool, optional): Should we store the alias in the cache? Defaults to True.

        Returns:
            Alias: The refreshed alias
        """
        assert self._lock is not None, (
            "You need to enter the context first before calling this function"
        )
        async with self._lock:
            if not self.loaded_fakts:
                try:
                    await self.aload()
                except Exception as e:
                    logger.error(e, exc_info=True)
                    raise e

        assert self.loaded_fakts, "No fakts loaded yet. Please call load() first."
        assert service_name in self.loaded_fakts.instances, (
            f"Service {service_name} not found in loaded fakts. Available services: {', '.join(self.loaded_fakts.instances.keys())}"
        )

        service_instance = self.loaded_fakts.instances[service_name]

        for alias in service_instance.aliases:
            if omit_challenge:
                # If we omit the challenge, we just return the alias
                self.alias_map[service_name] = alias
                return alias

            try:
                challenge_ok = await asyncio.wait_for(
                    self.achallenge_alias(alias), timeout=3
                )
                if challenge_ok:
                    self.alias_map[service_name] = alias
                    return alias
            except asyncio.TimeoutError as e:
                logger.error(
                    f"Timeout while challenging alias {alias} for service {service_name}.",
                    exc_info=True,
                )
                continue

        raise GroupNotFound(
            f"Could not find a valid alias for service {service_name}. Available aliases: {', '.join([str(alias.challenge_path) for alias in service_instance.aliases])} all failed to challenge."
        )

    async def aget_alias(
        self,
        fakts_key: Optional[str] = None,
        omit_challenge: bool = False,
        cache: bool = True,
        store: bool = True,
    ) -> Alias:
        """Get Fakt Value (async)

        Gets the currently active configuration for the group_name, by loading it from
        the grant if it is not already loaded.

        Steps:
            1. Acquire lock
            2. If not yet loaded and auto_load is True, load
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            auto_load (bool, optional): Should we autoload the configuration
                                        if nothing has been set? Defaults to True.
            force_refresh (bool, optional): Should we force a refresh of the grants.
                                            Grants can decide their own refresh logic?
                                            Defaults to False.

        Returns:
            dict: The active fakts
        """
        assert self._alias_lock is not None, (
            "You need to enter the context first before calling this function"
        )
        async with self._alias_lock:
            if fakts_key not in self.alias_map:
                # If we don't have the alias in the map, we need to refresh it
                try:
                    await self.arefresh_alias(fakts_key, omit_challenge=omit_challenge)
                except GroupNotFound as e:
                    logger.error(e, exc_info=True)
                    raise e

        assert fakts_key in self.alias_map, (
            f"Alias for key {fakts_key} not found in alias map. Available aliases: {', '.join(self.alias_map.keys())}"
        )
        return self.alias_map[fakts_key]

    def get_alias(
        self,
        fakts_key: Optional[str] = None,
        cache: bool = True,
        omit_challenge: bool = False,
        store: bool = True,
    ) -> Alias:
        """Get Fakt Value (sync)

        Gets the currently active configuration for the group_name, by loading it from
        the grant if it is not already loaded.

        Steps:
            1. Acquire lock
            2. If not yet loaded and auto_load is True, load
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            auto_load (bool, optional): Should we autoload the configuration
                                        if nothing has been set? Defaults to True.
            force_refresh (bool, optional): Should we force a refresh of the grants.
                                            Grants can decide their own refresh logic?
                                            Defaults to False.

        Returns:
            dict: The active fakts
        """
        return unkoil(
            self.aget_alias,
            fakts_key,
            cache=cache,
            store=store,
            omit_challenge=omit_challenge,
        )

    def get_token(self) -> str:
        """Get Authentikation Token for a service (sync)

        This method will return the currently loaded token, or refresh it if it is not
        loaded yet. It will raise an exception if the token could not be loaded.

        Returns:
            str: The currently loaded token
        """
        return unkoil(self.aget_token)

    async def __aenter__(self) -> "Fakts":
        """Enter the context manager

        This method will set the current fakts context variable to itself,
        and create locks, to make sure that only one fakt request is
        processed at a time.
        """

        current_fakts_next.set(
            self
        )  # TODO: We should set tokens, but depending on async/sync this is shit
        self._lock = asyncio.Lock()
        self._token_lock = asyncio.Lock()
        self._alias_lock = asyncio.Lock()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        """Exit the context manager and clean up"""
        current_fakts_next.set(
            None
        )  # TODO: And here we should reset, but can't because of koil unsafe thread

    def _repr_html_inline_(self) -> str:
        """(Internal) HTML representation for jupyter"""
        return f"<table><tr><td>grant</td><td>{self.grant.__class__.__name__}</td></tr></table>"


def get_current_fakts_next() -> Fakts:
    """Get the current fakts instance

    This method will return the current fakts instance, or raise an
    exception if no fakts instance is set.

    Returns
    -------
    Fakts
        The current fakts instance
    """
    fakts = current_fakts_next.get()

    if fakts is None:
        raise NoFaktsFound("No fakts instance set in this context")

    return fakts



from fakts_next.grants.errors import GrantError


class RemoteGrantError(GrantError):
    """Base class for all remotegrant errors"""

    pass


class ClaimError(RemoteGrantError):
    """An error that occurs when claiming the configuration from the endpoint"""

    pass


class DiscoveryError(RemoteGrantError):
    """An error that occurs when discovering the endpoint"""

    pass


class DemandError(RemoteGrantError):
    """An error that occurs when demanding the token from the endpoint"""

    pass



from typing import Dict, Optional
from fakts_next.models import ActiveFakts
from fakts_next.protocols import FaktsCache


class NoCache(FaktsCache):
    """A cache implementation that does not store any data."""

    async def aload(self) -> Optional[ActiveFakts]:
        """Loads the configuration from the grant

        It will try to load the configuration from the cache file.
        If the cache is expired, or the hash value is different from
        the one in the cache, it will load the grant again.

        Parameters
        ----------
        request : FaktsRequest
            The request object that may contain additional information needed for loading the configuration.

        Returns
        -------
        dict
            The configuration loaded from the grant.


        """

        return None

    async def aset(self, value: ActiveFakts):
        """Refreshes the configuration from the grant

        This function is used to refresh the configuration from the grant.
        This is used to refresh the configuration from the grant, and should
        be used to refresh the configuration from the grant.

        The request object is used to pass information
        """

        pass

    async def areset(self):
        """Resets the cache

        This function is used to reset the cache.
        This is used to reset the cache, and should
        be used to reset the cache.

        The request object is used to pass information
        """

        pass
